<?xml version="1.0" encoding="UTF-8"?>
<tileset name="tile_image_not_found_in_tileset" tilewidth="106" tileheight="72" tilecount="1" columns="0">
 <tile id="0">
  <image width="106" height="72" source="../../examples/hexmini_typo.png"/>
 </tile>
</tileset>
