<?xml version="1.0" encoding="UTF-8"?>
<tileset version="1.5" tiledversion="1.5.0" name="GrassAndWater" tilewidth="64" tileheight="64" tilecount="24" columns="4">
 <grid orientation="isometric" width="64" height="32"/>
 <image source="grass_and_water.png" width="256" height="384"/>
 <wangsets>
  <wangset name="Grass and Water" type="corner" tile="6">
   <wangcolor name="grass" color="#72ff62" tile="0" probability="1"/>
   <wangcolor name="water" color="#52c9ff" tile="23" probability="1"/>
   <wangtile tileid="0" wangid="0,1,0,1,0,1,0,1"/>
   <wangtile tileid="1" wangid="0,1,0,1,0,1,0,1"/>
   <wangtile tileid="2" wangid="0,1,0,1,0,1,0,1"/>
   <wangtile tileid="3" wangid="0,1,0,1,0,1,0,1"/>
   <wangtile tileid="4" wangid="0,1,0,2,0,1,0,1"/>
   <wangtile tileid="5" wangid="0,1,0,1,0,2,0,1"/>
   <wangtile tileid="6" wangid="0,1,0,1,0,1,0,2"/>
   <wangtile tileid="7" wangid="0,2,0,1,0,1,0,1"/>
   <wangtile tileid="8" wangid="0,2,0,2,0,2,0,1"/>
   <wangtile tileid="9" wangid="0,1,0,2,0,2,0,2"/>
   <wangtile tileid="10" wangid="0,2,0,1,0,2,0,2"/>
   <wangtile tileid="11" wangid="0,2,0,2,0,1,0,2"/>
   <wangtile tileid="12" wangid="0,1,0,2,0,2,0,1"/>
   <wangtile tileid="13" wangid="0,1,0,1,0,2,0,2"/>
   <wangtile tileid="14" wangid="0,2,0,1,0,1,0,2"/>
   <wangtile tileid="15" wangid="0,2,0,2,0,1,0,1"/>
   <wangtile tileid="16" wangid="0,1,0,2,0,2,0,1"/>
   <wangtile tileid="17" wangid="0,1,0,1,0,2,0,2"/>
   <wangtile tileid="18" wangid="0,2,0,1,0,1,0,2"/>
   <wangtile tileid="19" wangid="0,2,0,2,0,1,0,1"/>
   <wangtile tileid="20" wangid="0,2,0,1,0,2,0,1"/>
   <wangtile tileid="21" wangid="0,1,0,2,0,1,0,2"/>
   <wangtile tileid="22" wangid="0,2,0,2,0,2,0,2"/>
   <wangtile tileid="23" wangid="0,2,0,2,0,2,0,2"/>
  </wangset>
 </wangsets>
</tileset>
