.figure img {
    padding: 4px;
    background-color: #fff;
    border: 1px solid #ddd;
    border-radius: 4px;
    margin: 0px auto 5px auto;
}

.future .topic-title {
    border-bottom: 2px solid rgb(255, 148, 90);
    padding-bottom: 4px;
    color: rgb(254, 89, 0);

    /* make it a header */
    font-size: 150%;
    margin-top: 0;
    font-weight: 700;
    font-family: "Roboto Slab","ff-tisa-web-pro","Georgia",Arial,sans-serif;
}
.future a {
    color: rgb(254, 89, 0);
}

.new {
    font-weight: bold;
    font-size: 80%;
    background-color: rgb(254, 89, 0);
    border: 1px solid rgb(254, 89, 0);
    color: white;
    display: block;
    border-radius: 4px;
    float: right;
    padding: 3px;
    margin: 4px 0px 10px 10px;
}

.new-prev {
    background-color: unset;
    color: gray;
    border-color: gray;
}

/* override table no-wrap */
.wy-table-responsive table td {
    white-space: normal;
}

/* override table middle align */
.rst-content table.docutils td {
    vertical-align: top;
}

/* override too much space between line blocks */
.rst-content .line-block {
    margin-bottom: 0.5em;
}

.wy-side-nav-search {
    margin: 0px;        /* override unfitting bottom margin */
    width: auto;        /* override fixed 300px width */
    text-align: left;   /* override center alignment */
}

.wy-nav-top {
    display: block;     /* override hiding on wide views */
    padding: 0px;       /* override unwanted padding */
    line-height: normal;/* override 50px */
    max-width: 800px;   /* matches content max width */
}

.wy-nav-top i {
    float: none;        /* override left float */
    vertical-align: middle;
    padding-top: 0px;   /* override inheriting padding */
    visibility: hidden;
}

@media screen and (max-width: 768px) {
    .wy-nav-top i {
        visibility: visible;
    }
}

.wy-side-nav-search input[type=text] {
    vertical-align: middle !important; /* override top alignment */
}

.wy-side-nav-search > a {
    margin-left: 5px;
    margin-bottom: 0px; /* override unwanted margin */
}

/* more generic rule (theme only applies to div.version) */
.wy-side-nav-search > .version {
    color: rgba(255,255,255,0.3);
}
.wy-side-nav-search > div {
    float: right;
}
