/*
 * addpropertydialog.h
 * Copyright 2015, CaptainFrog <jwilliam.perreault@gmail.com>
 * Copyright 2016, Thorbjørn Lindeijer <thorbjorn@lindeijer.nl>
 *
 * This file is part of Tiled.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <QDialog>
#include <QVariant>

namespace Tiled {
class ClassPropertyType;
}

namespace Ui {
class AddPropertyDialog;
}

class AddPropertyDialog : public QDialog
{
    Q_OBJECT

public:
    explicit AddPropertyDialog(QWidget *parent = nullptr);
    AddPropertyDialog(const Tiled::ClassPropertyType *parentClassType, QWidget *parent = nullptr);
    ~AddPropertyDialog() override;

    QString propertyName() const;
    QVariant propertyValue() const;

private:
    void initialize(const Tiled::ClassPropertyType *parentClassType);

    void nameChanged(const QString &text);
    void typeChanged(const QString &text);

    Ui::AddPropertyDialog *mUi;
};
